﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.ComponentModel.Composition;
using DotSpatial.Controls;
using DotSpatial.Topology;
using DotSpatial.Serialization;
using DotSpatial.Data;
using DotSpatial.Symbology;

namespace MES_Wind
{
    public partial class frmMain : Form
    {
        [Export("Shell", typeof(ContainerControl))]
        private static ContainerControl Shell;
        #region "Control variables"
         double distThreshold = 500;
        #endregion

        public frmMain()
        {

            InitializeComponent();
            if (DesignMode) return;
            Shell = this;
            appManager1.LoadExtensions();
        }
        #region
        /// <summary>
        /// This function is used to check if the segment of powerline is broken. 
        /// Based on the given line segment's start and endpoint, it divides the line on
        /// suitable subsegments defined by threshold length - dThreshold
        /// </summary>
        /// <param name="startX">Line segement's start X point</param>
        /// <param name="startY">Line segement's start Y point</param>
        /// <param name="endX">Line segement's end X point</param>
        /// <param name="endY">Line segement's end Y point</param>
        /// <param name="dThreshold">Line segment length threshold</param>
        /// <param name="Uwind_raster">Raster Layer for progmostic wind X coord</param>
        /// <param name="Vwind_raster">Raster Layer for progmostic wind Y coord</param>
        /// <returns>List of booleans with coordinates if any of them are true, line is broken</returns>
        /// <remarks></remarks>
#endregion
        List<PRMLibrary.Coordinate> DotspLinestringToPrm(LineString dtlinestr)
        {
            List<PRMLibrary.Coordinate> prmlinestring = new List<PRMLibrary.Coordinate>();
            IList<Coordinate> linepoints = dtlinestr.Coordinates;
            foreach(Coordinate linepoint in linepoints)
            {
                PRMLibrary.Coordinate point = new PRMLibrary.Coordinate(linepoint.X, linepoint.Y);
                prmlinestring.Add(point);
            }
            return prmlinestring;
        }
        PRMLibrary.Coordinate DotspPointToPRM(IPoint ftpoint)
        {
            PRMLibrary.Coordinate prpoint = new PRMLibrary.Coordinate();
            prpoint.X = ftpoint.X;
            prpoint.Y = ftpoint.Y;
            return prpoint;
        }
        public List<CheckPoint> CalcBrkPoint(double startX, double startY, double endX, double endY, double dThreshold, IMapRasterLayer Uwind_raster, IMapRasterLayer Vwind_raster, IMapRasterLayer clim_layer, double h)
        {
            List<CheckPoint> lineCheckPoint = new List<CheckPoint>();
            double uwind = 0;
            double vwind = 0;
            double umod = 0;
            double sinwind = 0;
            double anglewind = 0;
            double angleline = Math.Atan2((endY - startY),(endX - startX));
            double climwind = 0;
            double distance = Math.Sqrt((endX - startX) * (endX - startX) + (endY - startY) * (endY - startY));
            double distpropD = distance / dThreshold;
            int distpropI = Convert.ToInt32(distpropD);
            double curX = startX;
            double curY = startY;
            CheckPoint chkpnt = new CheckPoint();
            if (distpropI > 1)
            {
                double constXdiff = (endX - startX) / distpropI;
                double constYdiff = (endY - startY) / distpropI;
                for (int j = 1; j < distpropI + 1; j++)
                {
                    if (j == 1)
                    {
                        curX = startX + constXdiff / 2;
                        curY = startY + constXdiff / 2; 
                    }
                    else
                    {
                        curX = curX + constXdiff;
                        curY = curY + constYdiff;
                    }
                    Coordinate coords = new Coordinate(curX,curY);
                    uwind = interpol(coords, Uwind_raster);
                    vwind = interpol(coords, Vwind_raster);
                    climwind = interpol(coords, clim_layer);
                    umod = Math.Sqrt(uwind*uwind + vwind*vwind);
                    anglewind = Math.Atan2(vwind,uwind) - angleline;
                    sinwind = Math.Sin(anglewind);
                    double C_height = 1.0;
                    if (umod < 20)
                    { //wind is too low 
                        chkpnt.Ifbroken = false;
                    }
                    else
                    {    // calculate prognostic and threshold windstress
                        double p1 = -0.00078501;
                        double p2 = 0.13431;
                        double p3 = -2.11112;
                        double p4 = 19.548;
                        double qpr = p1 * umod * umod * umod + p2 * umod * umod + p3 * umod + p4;
                        double qcl = p1 * climwind * climwind * climwind + p2 * climwind * climwind + p3 * climwind + p4;
                        double Ppr = qpr * C_height * sinwind * sinwind;
                        double Pcl = qcl * C_height * 1.0;
                        if (Ppr >= Pcl)
                        {
                            chkpnt.Ifbroken = true;
                        }
                        else
                        {
                            chkpnt.Ifbroken = false;
                        }
                    }
                    chkpnt.X = curX;
                    chkpnt.Y = curY;
                   
                    lineCheckPoint.Add(chkpnt);
                }
            }
            else
            {
                 
            }
            return lineCheckPoint;
        }
        public IFeatureSet brokenpoints(List<CheckPoint> chklist)
        {
            IFeatureSet points = new FeatureSet(FeatureType.Point);
            Coordinate chkcords = new Coordinate();
            return points;
        }
        public IFeatureSet main_layer_function(IMapLineLayer pwlines, IFeatureSet pwlineSet, IMapRasterLayer u_raster,
            IMapRasterLayer v_raster, IMapRasterLayer clim15_raster, IMapRasterLayer clim10_raster, IMapRasterLayer clim5_raster)
        {
            IFeatureSet result_layer = new FeatureSet(FeatureType.Line);
            List<CheckPoint> fullCheckList = new List<CheckPoint>();
            foreach (IFeature feature in pwlineSet.Features)
            {
                List<CheckPoint> lineCheckList = new List<CheckPoint>();
                //get associated attributes
                DataRow featureData = feature.DataRow;
                int id = int.Parse(featureData["Shape_ID"].ToString());
                int year = int.Parse(featureData["Year"].ToString());
                double height = double.Parse(featureData["HeightOffs"].ToString());
                int power = int.Parse(featureData["Power"].ToString());
                LineString linestr = feature.BasicGeometry as LineString;
                if (linestr != null)
                { // case if powerline consists of one line
                  // get coordinates list
                    IList<Coordinate> points = linestr.Coordinates;


                    // cycle throw all points in line
                    for (int i = 1; i < points.Count; i++)
                    {
                        List<CheckPoint> segmentCheckList = new List<CheckPoint>();
                        double x1 = points[i - 1].X;
                        double y1 = points[i - 1].Y;
                        double x2 = points[i].X;
                        double y2 = points[i].Y;
                        if (power > 5 && power < 330)
                        {
                            segmentCheckList = CalcBrkPoint(x1, y1, x2, y2, distThreshold, u_raster, v_raster, clim10_raster, height);
                        }
                        else
                        {
                            if (power < 5)
                            {
                                segmentCheckList = CalcBrkPoint(x1, y1, x2, y2, distThreshold, u_raster, v_raster, clim5_raster, height);
                            }
                            else
                            {
                                segmentCheckList = CalcBrkPoint(x1, y1, x2, y2, distThreshold, u_raster, v_raster, clim15_raster, height);
                            }
                            }
                        lineCheckList.AddRange(segmentCheckList);
                    }
                    bool linechek = false;
                    foreach (CheckPoint chkpnt in lineCheckList)
                    {
                        if (chkpnt.Ifbroken == true)
                        { linechek = true; }
                    }
                    if (linechek == true)
                    {
                        IFeature lineFeature = result_layer.AddFeature(linestr);
                    }
                    fullCheckList.AddRange(lineCheckList);

                }
                else {
                    MessageBox.Show("shit you trying to feed it is not a simple line");
                }
            }

            return result_layer;
        }
        public double interpol(Coordinate coords, IMapRasterLayer raster)
        {
            const bool normalX = true;
            const bool normalY = false;
            RcIndex rc = raster.DataSet.Bounds.ProjToCell(coords);
            Coordinate center = raster.DataSet.Bounds.CellCenter_ToProj(rc.Row, rc.Column);
            double xDiff = coords.X - center.X;
            double yDiff = coords.Y - center.Y;
            //calculate second index
            int row2, col2;
            if ((xDiff >= 0 && normalX ) || (!normalX && xDiff < 0))
            {
                row2 = rc.Row >= raster.DataSet.EndRow ? rc.Row - 1 : rc.Row + 1;
            }
            else
            {
                row2 = rc.Row > 0 ? rc.Row - 1 : rc.Row + 1;
            }
            if ( (yDiff >= 0 && normalY) || (!normalY && yDiff < 0))
            {
                col2 = rc.Column >= raster.DataSet.EndColumn ? rc.Column - 1 : rc.Column + 1;
            }
            else
            {
                col2 = rc.Column > 0 ? rc.Column - 1 : rc.Column + 1;
            }
            // indexes and values at bounds
            RcIndex rcBotLeft  = new RcIndex(Math.Min(row2, rc.Row), Math.Min(col2, rc.Column));
            RcIndex rcBotRight = new RcIndex(Math.Max(row2, rc.Row), Math.Min(col2, rc.Column));
            RcIndex rcTopLeft  = new RcIndex(Math.Min(row2, rc.Row), Math.Max(col2, rc.Column));
            RcIndex rcTopRight = new RcIndex(Math.Max(row2, rc.Row), Math.Max(col2, rc.Column));
            double valBotLeft  = raster.DataSet.Value[rcBotLeft.Row, rcBotLeft.Column];
            double valBotRight = raster.DataSet.Value[rcBotRight.Row, rcBotRight.Column];
            double valTopLeft  = raster.DataSet.Value[rcTopLeft.Row, rcTopLeft.Column];
            double valTopRight = raster.DataSet.Value[rcTopRight.Row, rcTopRight.Column];
            Coordinate origin = raster.DataSet.CellToProj(rcBotLeft.Row, rcBotLeft.Column);
            //Coordinate last = raster.DataSet.CellToProj(rcTopRight.Row, rcTopRight.Column);//test only
            // sizes for cell
            double hx = raster.DataSet.Bounds.CellWidth;
            double hy = raster.DataSet.Bounds.CellHeight;
            // coefficients
            double px = (coords.X - origin.X) / hx;
            double py = (coords.Y - origin.Y) / hy;
            // inverse directions
            px *= normalX ? 1 : -1;
            py *= normalY ? 1 : -1;
            // interpolation
            double top = (1 - px) * valTopLeft + px * valTopRight;
            double bot = (1 - px) * valBotLeft + px * valBotRight;
            double rval = (1 - py) * bot + py * top;
            return rval;
        }

        // so, we start at source power Point; actually call this foreach source point
        public void checkPowerPoint(PowerPointObject sourcePoint)
        {
            if (!sourcePoint.powerIsON) { 
                // so we have source without power
                MessageBox.Show("checkPowerPoint called with disabled source point");
                return;
            }
            // looking for unbroken lines and switch power for this point
            foreach (PowerLineObject line in sourcePoint.lines) {
                if (!line.broken && !line.toPoint.powerIsON) {
                    // switch power and check on subpoints.
                    line.toPoint.powerIsON = true;
                    checkPowerPoint(line.toPoint);
                }
                // else we have broken line or already switched power point
            }
        }

        // find all disabled power points
        public List<PowerPointObject> findDisabledPoints(List<PowerPointObject> pPoints)
        {
            List<PowerPointObject> result = new List<PowerPointObject>();
            foreach (PowerPointObject point in pPoints) {
                if (!point.powerIsON) {
                    // if disabled, we add it
                    result.Add(point);
                }
            }
            return result;
        }
        private void bntLoadWindX_Click(object sender, EventArgs e)
        {
            map1.AddLayer();
            map1.ZoomToMaxExtent();

        }

        private void btnLoadWindY_Click(object sender, EventArgs e)
        {
            string curDir = Environment.CurrentDirectory;
            string path_to_tests = "\\MES_test_new\\";
            string file_path = curDir + path_to_tests;

            map1.AddLayer(file_path+"u_test.asc");
            map1.AddLayer(file_path + "v_test.asc");
            map1.AddLayer(file_path + "clim5_test.asc");
            map1.AddLayer(file_path + "clim10_test.asc");
            map1.AddLayer(file_path + "clim15_test.asc");
            map1.AddLayer(file_path + "powerlines.shp");
            map1.AddLayer(file_path + "powerstations.shp");
            map1.ZoomToMaxExtent();
        }

        private void btnViewAttributeTable_Click(object sender, EventArgs e)
        {
            //Declare a datatable
            DataTable dt = null;

            if (map1.GetLineLayers().Count() > 0)
            {
                MapLineLayer pwlLayer = default(MapLineLayer);

                pwlLayer = (MapLineLayer)map1.GetLineLayers()[0];

                if (pwlLayer == null)
                {
                    MessageBox.Show("The layer is not a line layer.");
                }
                else
                {
                    //Get the shapefile's attribute table to our datatable dt
                    dt = pwlLayer.DataSet.DataTable;

                    //Set the datagridview datasource from datatable dt
                    dgvAttributeTable.DataSource = dt;
                }
            }
            else
            {
                MessageBox.Show("Please add a layer to the map.");
            }

        

    }

        private void btnCalcStress_Click(object sender, EventArgs e)
        {
            try
            {
                //extract prognostic u layer
                IMapRasterLayer u_rasterLayer = default(IMapRasterLayer);
                IMapRasterLayer v_rasterLayer = default(IMapRasterLayer);
                IMapRasterLayer clim5_rasterLayer = default(IMapRasterLayer);
                IMapRasterLayer clim10_rasterLayer = default(IMapRasterLayer);
                IMapRasterLayer clim15_rasterLayer = default(IMapRasterLayer);

                if (map1.GetRasterLayers().Count() == 1)
                {
                    MessageBox.Show("Please add a raster layer");
                    return;
                }

                //this makes raster layers list in map1 ordered!
                u_rasterLayer = map1.GetRasterLayers()[0];
                v_rasterLayer = map1.GetRasterLayers()[1];
                clim5_rasterLayer = map1.GetRasterLayers()[2];
                clim10_rasterLayer = map1.GetRasterLayers()[3];
                clim15_rasterLayer = map1.GetRasterLayers()[4];

                //get the powerline  line layer
                IMapLineLayer pwlLayer = default(IMapLineLayer);
                if (map1.GetLineLayers().Count() == 0)
                {
                    MessageBox.Show("Please add powerline shapefile");
                    return;
                }
                pwlLayer = map1.GetLineLayers()[0];
                //copy line layer FeatureSet
                IFeatureSet pwlineSet = pwlLayer.DataSet;
                //get the powerstations layer 
                IMapPointLayer pwstLayer = default(IMapPointLayer);
                pwstLayer = map1.GetPointLayers()[0];
                // copy point layer FeatureSet
                IFeatureSet pwpointsSet = pwstLayer.DataSet;
                //Start to cast raster to PRM classes
                // prognostic wind massives first
                List<List<PRMLibrary.PrognosticCell>> prog_wind = new List<List<PRMLibrary.PrognosticCell>>();

                PRMLibrary.Coordinate dummy_coord = new PRMLibrary.Coordinate();
                Coordinate dummy_rcoord = new Coordinate();
                PRMLibrary.PrognosticCell dummy_prognostic = new PRMLibrary.PrognosticCell(dummy_coord, 0, 0);
                int rcount_prog = u_rasterLayer.DataSet.NumRows;
                int ccount_prog = u_rasterLayer.DataSet.NumColumns;
                for (int i =0; i< ccount_prog; i++)
                {
                    List<PRMLibrary.PrognosticCell> prog_wind_row = new List<PRMLibrary.PrognosticCell>();
                    for (int j =0; j< rcount_prog; j++ )
                    {
                        dummy_prognostic.velocityX = u_rasterLayer.DataSet.Value[j, i];
                        dummy_prognostic.velocityY = v_rasterLayer.DataSet.Value[j, i];
                        dummy_rcoord = u_rasterLayer.Bounds.CellCenter_ToProj(j,i);
                        dummy_coord.X = dummy_rcoord.X;
                        dummy_coord.Y = dummy_rcoord.Y;
                        dummy_prognostic.coords = dummy_coord;
                        prog_wind_row.Add(dummy_prognostic);                        
                    }
                    prog_wind.Add(prog_wind_row);
                    //prog_wind_row.Clear();
                }
                //Get cell info and Affine transform coefficients from prognostic wind rasters
                PRMLibrary.CellSize prog_cell = new PRMLibrary.CellSize(u_rasterLayer.DataSet.CellWidth, u_rasterLayer.DataSet.CellHeight);
                double[] prog_aff = u_rasterLayer.Bounds.AffineCoefficients;
                //Now we create climate raster class
                List<List<PRMLibrary.ClimateCell>> clim_wind = new List<List<PRMLibrary.ClimateCell>>();
                PRMLibrary.ClimateCell dummy_clim = new PRMLibrary.ClimateCell(dummy_coord, 0, 0, 0);
                int rcount_clim = clim5_rasterLayer.DataSet.NumRows;
                int ccount_clim = clim5_rasterLayer.DataSet.NumColumns;
                for (int i = 0; i < ccount_clim; i++)
                {
                    List<PRMLibrary.ClimateCell> clim_wind_row = new List<PRMLibrary.ClimateCell>();
                    for (int j = 0; j < rcount_clim; j++)
                    {
                        dummy_clim.wind5 = clim5_rasterLayer.DataSet.Value[j, i];
                        dummy_clim.wind10 = clim10_rasterLayer.DataSet.Value[j, i] ; // Important:  substract 3 to this to get powerlines broken
                        dummy_clim.wind15 = clim15_rasterLayer.DataSet.Value[j, i];
                        dummy_rcoord = clim15_rasterLayer.Bounds.CellCenter_ToProj(j,i);
                        dummy_coord.X = dummy_rcoord.X;
                        dummy_coord.Y = dummy_rcoord.Y;
                        dummy_clim.coords = dummy_coord;
                        clim_wind_row.Add(dummy_clim);
                    }
                    clim_wind.Add(clim_wind_row);
                    //clim_wind_row.Clear();
                }
                //Get cell info and affine transform coeff from climate rasters
                PRMLibrary.CellSize clim_cell = new PRMLibrary.CellSize(clim5_rasterLayer.DataSet.CellWidth, clim5_rasterLayer.DataSet.CellHeight);
                double[] clim_aff = clim5_rasterLayer.Bounds.AffineCoefficients;
                // create PRM_line list to pass to PRM_wind from loaded line layer
                List<PRMLibrary.Powerline> powerlinesToPRM = new List<PRMLibrary.Powerline>();
                foreach (IFeature feature in pwlineSet.Features)
                {
                    PRMLibrary.Powerline dummyline = new PRMLibrary.Powerline();
                    DataRow featureData = feature.DataRow;
                    dummyline.identifier = feature.Fid;
                    dummyline.year = int.Parse(featureData["Year"].ToString());
                    dummyline.height = double.Parse(featureData["HeightOffs"].ToString());
                    dummyline.power = int.Parse(featureData["Power"].ToString());
                    dummyline.pointFromID = int.Parse(featureData["PointFrom"].ToString());
                    dummyline.pointToID = int.Parse(featureData["PointTo"].ToString());
                    LineString ftline = feature.BasicGeometry as LineString;
                    dummyline.coords = DotspLinestringToPrm(ftline);
                    powerlinesToPRM.Add(dummyline);
                }
                //create PRM_station list to pass to PRM_wind from loaded point layer
                List<PRMLibrary.PowerStation> powerpointsToPRM = new List<PRMLibrary.PowerStation>();
                foreach (IFeature ftpoint in pwpointsSet.Features)
                {
                    PRMLibrary.PowerStation dummystation = new PRMLibrary.PowerStation();
                    DataRow featureData = ftpoint.DataRow;
                    dummystation.identifier = ftpoint.Fid;
                    dummystation.name = featureData["Name"].ToString();
                    dummystation.power = int.Parse(featureData["Power"].ToString());
                    int issr = int.Parse(featureData["IsSource"].ToString());
                    if (issr == 0)
                    {
                        dummystation.issource = false;
                    }
                    else
                    {
                        if (issr == 1)
                        {
                            dummystation.issource = true;
                        }
                        else
                        {
                            MessageBox.Show("Some enteties in Column IsSource of powerstation datatable are not strict ones or zeros");
                        }
                    }
                    dummystation.type = featureData["Type"].ToString();
                    IPoint ftpt = ftpoint.BasicGeometry as IPoint;
                    dummystation.coords = DotspPointToPRM(ftpt);
                    powerpointsToPRM.Add(dummystation);

                }
                //Create a PRM_wind class and add all the properties from above
                PRMLibrary.Module prmwind = new PRMLibrary.Module();
                prmwind.powerLines = powerlinesToPRM;
                prmwind.powerStations = powerpointsToPRM;
                prmwind.prognosticCells = prog_wind;
                prmwind.prognosticCellSize = prog_cell;
                prmwind.prognosticAffineCoefficients = prog_aff;
                prmwind.climateCells = clim_wind;
                prmwind.climateCellSize = clim_cell;
                prmwind.climateAffineCoefficients = clim_aff;
                //Calculate which lines are broken
                List<PRMLibrary.Powerline> prmBrokenLines = prmwind.brokenPowerLinesAfterCheck();
                prmwind.checkPower();
                // new FeatureSet for resulting broken powerlines
                //IFeatureSet brklineSet = new FeatureSet(FeatureType.Line);
                //DataTable dt = pwlineSet.DataTable;
                IFeatureSet brk_info = new FeatureSet(FeatureType.Line);
                brk_info = main_layer_function(pwlLayer, pwlineSet, u_rasterLayer, v_rasterLayer, clim15_rasterLayer, clim10_rasterLayer, clim5_rasterLayer);

                IMapLineLayer brk_info_layer = (MapLineLayer)map1.Layers.Add(brk_info);
                LineSymbolizer symbol = new LineSymbolizer(Color.Red, 3);
                brk_info_layer.Symbolizer = symbol;
                brk_info_layer.LegendText = "Broken powerlines";

            }
            catch (Exception ex)
            {
                MessageBox.Show("Something went wrong");
            }

        }
    }
    public class CheckPoint
    {
        public double X;
        public double Y;
        public bool Ifbroken;
    }

}
